// This version of binary trees stores a root pointer and a size in a Tree
// struct. This way, we can find out the size of a tree without counting it
// every time. However, what keeps the size in sync with the actual tree?
// (Nothing but careful programming.) See bintree3.h for a version that
// ensures that we construct trees properly and makes the operations members
// of the struct.
#pragma once

#include <memory>
#include <stdexcept>
#include <string>
#include <vector>

namespace bintree {

// A Tree now has a shared_ptr to a node and keeps track of its size.
struct Tree {
    struct Node;
    using link_t = std::shared_ptr<Node>;

    link_t root;
    size_t size;
};

// An exception thrown by lookup(Tree, const std::string&) when the key isn't
// found.
struct Not_found : std::runtime_error
{
    Not_found(const std::string& key) :
            runtime_error("key not found: " + key) { }
};

// The number of elements in the tree.
size_t size(const Tree&);

// Is the key present in the tree?
bool member(const Tree&, const std::string&);

// Returns the value associated with the key in the tree, or throws Not_found
// if the key is absent.
unsigned lookup(const Tree&, const std::string&);

// Inserts or updates the value associated with a key in the tree.
void insert(Tree&, const std::string&, unsigned);

// Returns all keys in order as a vector.
std::vector<std::string> get_keys(const Tree&);

}